__version__ = "0.0.0"

import os
import re
import sys
import time
import logging
import requests
import traceback
import subprocess
from logging.handlers import RotatingFileHandler

import pyaudio
import timeout_decorator
from six.moves import queue
from google.cloud import speech
from google.cloud import texttospeech
from google.cloud.speech import enums
from google.cloud.speech import types
from google.api_core.exceptions import DeadlineExceeded

from dbmodel import DbModel as Names


class RaspberryClient:

    def __init__(self, user_id):
        self.USER_ID = user_id

        self.HOST = sys.argv[1]
        self.LOG_FILENAME = "./logs/" + self.USER_ID + "_debug.txt"
        self.FORMATTER = logging.Formatter(
            '[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(funcName)s -  %(message)s')
        self.sentence_splitter = re.compile(r'.*?[.!\?]')

        # Audio recording parameters
        self.RATE = 16000
        self.CHUNK = int(self.RATE / 10)  # 100ms

        # See http://g.co/cloud/speech/docs/languages
        # for a list of supported languages.
        self.language_code = 'cs-CZ'  # a BCP-47 language tag

        # Instantiates a client
        self.client = texttospeech.TextToSpeechClient()
        self.client_STT = speech.SpeechClient()

        self.voice = texttospeech.types.VoiceSelectionParams(
            language_code='cs-CZ',
            name="cs-CZ-Wavenet-A",
            ssml_gender=texttospeech.enums.SsmlVoiceGender.FEMALE)

        # Select the type of audio file you want returned
        self.audio_config = texttospeech.types.AudioConfig(
            audio_encoding=texttospeech.enums.AudioEncoding.MP3,
            pitch=0)

        self.logger = self.get_logger("myLogger")

    class MicrophoneStream(object):
        """Opens a recording stream as a generator yielding the audio chunks."""

        def __init__(self, rate, chunk):
            self._rate = rate
            self._chunk = chunk

            # Create a thread-safe buffer of audio data
            self._buff = queue.Queue()
            self.closed = True

        def __enter__(self):
            self._audio_interface = pyaudio.PyAudio()
            self._audio_stream = self._audio_interface.open(
                format=pyaudio.paInt16,
                # The API currently only supports 1-channel (mono) audio
                # https://goo.gl/z757pE
                channels=1, rate=self._rate,
                input=True, frames_per_buffer=self._chunk,
                # Run the audio stream asynchronously to fill the buffer object.
                # This is necessary so that the input device's buffer doesn't
                # overflow while the calling thread makes network requests, etc.
                stream_callback=self._fill_buffer,
            )

            self.closed = False

            return self

        def __exit__(self, type, value, traceback):
            self._audio_stream.stop_stream()
            self._audio_stream.close()
            self.closed = True
            # Signal the generator to terminate so that the client's
            # streaming_recognize method will not block the process termination.
            self._buff.put(None)
            self._audio_interface.terminate()

        def _fill_buffer(self, in_data, frame_count, time_info, status_flags):
            """Continuously collect data from the audio stream, into the buffer."""
            self._buff.put(in_data)
            return None, pyaudio.paContinue

        def generator(self):
            while not self.closed:
                # Use a blocking get() to ensure there's at least one chunk of
                # data, and stop iteration if the chunk is None, indicating the
                # end of the audio stream.
                chunk = self._buff.get()
                if chunk is None:
                    return
                data = [chunk]

                # Now consume whatever other data's still buffered.
                while True:
                    try:
                        chunk = self._buff.get(block=False)
                        if chunk is None:
                            return
                        data.append(chunk)
                    except queue.Empty:
                        break

                yield b''.join(data)

    def listen_print_loop(self, responses):
        best_alternative = "{{ Text nebyl rozpoznán. }}"
        best_confidence_score = 0
        was_silent = True
        for response in responses:
            for result in response.results:
                was_silent = False
                self.logger.info('Finished: {}'.format(result.is_final))
                self.logger.info('Stability: {}'.format(result.stability))
                alternatives = result.alternatives
                # The alternatives are ordered from most likely to least.
                for alternative in alternatives:
                    self.logger.info('Confidence: {}'.format(alternative.confidence))
                    self.logger.info(u'Transcript: {}'.format(alternative.transcript))
                    if alternative.confidence >= best_confidence_score:
                        best_alternative = alternative.transcript
                        best_confidence_score = alternative.confidence
                if result.is_final:
                    return best_alternative, best_confidence_score
            if was_silent:
                return best_alternative, best_confidence_score

    try:
        from gpiozero import LED, Button

        LED_PIN = 25
        led = LED(LED_PIN)
        button = Button(23)

    except ImportError:
        class FakeLED:

            def on(self):
                print("Led On")

            def off(self):
                print("Led Off")

        class FakeButton:
            def __init__(self):
                self.when_pressed = None

        led = FakeLED()
        button = FakeButton()

    def get_console_handler(self):
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(self.FORMATTER)
        return console_handler

    def get_file_handler(self):
        file_handler = RotatingFileHandler(self.LOG_FILENAME, maxBytes=10000000, backupCount=5)
        file_handler.setFormatter(self.FORMATTER)
        return file_handler

    def get_logger(self, logger_name):
        os.makedirs("temp", exist_ok=True)
        os.makedirs("logs", exist_ok=True)
        os.makedirs("logs/audio", exist_ok=True)

        logger = logging.getLogger(logger_name)
        logger.setLevel(logging.DEBUG)  # better to have too much log than not enough
        logger.addHandler(self.get_console_handler())
        logger.addHandler(self.get_file_handler())
        # with this pattern, it's rarely necessary to propagate the error up to parent
        logger.propagate = False
        return logger


    @timeout_decorator.timeout(8)
    def open_mic_stream(self, streaming_config):
        with self.MicrophoneStream(self.RATE, self.CHUNK) as stream:
            audio_generator = stream.generator()
            _requests = (types.StreamingRecognizeRequest(audio_content=content)
                         for content in audio_generator)

            responses = self.client_STT.streaming_recognize(streaming_config, _requests)

            text, confidence_score = self.listen_print_loop(responses)
            return text, confidence_score

    def listen(self, data):
        self.logger.info(data)

        text = "{{ Text nebyl rozpoznán. }}"
        confidence_score = 0.0

        self.led.on()

        config = types.RecognitionConfig(
            encoding=enums.RecognitionConfig.AudioEncoding.LINEAR16,
            sample_rate_hertz=self.RATE,
            language_code=self.language_code,
            max_alternatives=5,
            speech_contexts=[speech.types.SpeechContext(
                phrases=data[Names.EXPECTED_PHRASES] if Names.EXPECTED_PHRASES in data else [])]
        )
        streaming_config = types.StreamingRecognitionConfig(
            config=config,
            single_utterance=True,
            interim_results=True)
        try:
            text, confidence_score = self.open_mic_stream(streaming_config)
        except timeout_decorator.TimeoutError:
            text = "{{ Časové okno pro rozpoznávní bylo překročeno. }}"
            self.logger.error(traceback.format_exc())
        except DeadlineExceeded:
            text = "{{ Text rozpoznán s chybou. }}"
            self.logger.debug("google.api_core.exceptions.DeadlineExceeded")

        self.led.off()

        self.logger.info(text)
        self.logger.info(confidence_score)

        data[Names.PREVIOUS_MESSAGE] = text
        data[Names.CONFIDENCE_SCORE] = confidence_score

        if data[Names.INITIAL_PHASE]:
            r = requests.post(self.HOST + "/first_listened/", json=data, timeout=5)
        else:
            r = requests.post(self.HOST + "/listened/", json=data, timeout=5)

        self.logger.info(r.json())
        self.play(r.json())

    def tts_google(self, text):
        self.logger.info(text)
        synthesis_input = texttospeech.types.SynthesisInput(ssml=text)

        response = self.client.synthesize_speech(synthesis_input, self.voice, self.audio_config)

        src = "./temp/" + self.USER_ID + "_" + str(time.time()) + ".mp3"

        with open(src, "wb") as f:
            # Write the response to the output file.
            f.write(response.audio_content)

        subprocess.call(["mpg321", "--gain", "10", src])  # waiting to complete
        os.remove(src)

    def play(self, data):
        self.logger.info(data)

        self.tts_google(data[Names.MESSAGE])
        if data[Names.END] is False:
            self.listen(data)

    def force_add_task(self):
        try:
            self.led.on()
            r = requests.post(self.HOST + "/job/force/" + self.USER_ID, timeout=5)
            self.logger.info(r.json())
            self.led.off()
        except Exception as e:
            self.led.on()
            time.sleep(1)
            self.led.off()
            time.sleep(1)
            self.led.on()
            time.sleep(1)
            self.led.off()
            self.logger.error(e)

    def run(self):
        self.button.when_pressed = self.force_add_task

        # tts_google("<speak>Rozdělená verze 3</speak>")
        if "GOOGLE_APPLICATION_CREDENTIALS" not in os.environ \
                or not os.path.isfile(os.environ["GOOGLE_APPLICATION_CREDENTIALS"]):
            self.logger.error("You should set proper GOOGLE_APPLICATION_CREDENTIALS.")
        if os.path.isfile("/home/pi/cred.json"):
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = "/home/pi/cred.json"
        else:
            self.logger.error("Or at least place cred.json into /home/pi/ directory")
            return False

        if "DEBUG" in os.environ and int(os.environ["DEBUG"]) == 1:
            self.force_add_task()

        while True:
            try:
                r = requests.get(self.HOST + "/job/" + self.USER_ID, timeout=5)
                self.logger.info(r.json())
                if requests.get("https://promethist.ai/repository/pip/brainquist/raspberryclient/version.txt").text.strip() == __version__:
                    return True # should update
                elif Names.NO_TASK in r.json():
                    time.sleep(60)  # one minute
                else:
                    self.play(r.json())
            except (SystemExit, Exception):
                tb = traceback.format_exc()
                self.logger.error(tb)
                return False
